#include <linux/limits.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/wait.h>
#include <ctype.h>
#include <dirent.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <string.h>

#ifndef PORT
#define PORT 10000
#endif

#define MAXLINE 1024
#define BUFSZ	4096
#define IDSZ	16
#define MAXFILES 1000

char filenames[BUFSZ];
char secret[MAXLINE];

struct cmd {
	char *files[MAXFILES];
	char id[IDSZ];
	int nselected;
	int nvalid;
};

void fillsecret(char *);
void readfiles(struct cmd *c);
void selectfiles(struct cmd *c, const char *filter);

void execcmd(struct cmd *c)
{
	char *argv[MAXFILES + 1];
	int argc;

	for (argc = 0; argc < c->nselected; argc++)
		argv[argc] = c->files[argc];
	argv[argc] = NULL;
	execv(c->files[0], argv);
	perror(c->files[0]);
}

void child()
{
	char buf[MAXLINE];
	int i;
	size_t n;
	struct cmd c;

	clearenv();

	if (chdir("files") < 0) {
		perror("files");
		exit(1);
	}

	fillsecret(secret);
	readfiles(&c);

	printf("Frobnicate server v1.0\n");

	for (;;) {
		if (fgets(buf, MAXLINE, stdin) == NULL)
			break;
		n = strlen(buf);
		if (n < 2)
			continue;
		if (buf[n - 1] != '\n') {
			fprintf(stderr, "line too long, ignored\n");
			continue;
		}

		switch (buf[0]) {
		case 'l': // list files
			for (i = 1; i < c.nvalid; i++)
				printf("%c%s\n", i < c.nselected ? '*' : ' ', c.files[i]);
			break;
		case 'c': // cat files
			c.files[0] = "/bin/cat";
			execcmd(&c);
			break;
		case 'f': // frobnicate files
			c.files[0] = "../frobnicate";
			printf("id: ");
			if (fgets(buf, MAXLINE, stdin) == NULL)
				break;
			n = strlen(buf);
			if (buf[n - 1] != '\n') {
				fprintf(stderr, "id too long, aborted\n");
				break;
			}
			buf[n - 1] = '\0';
			if (strlen(buf) > IDSZ) {
				fprintf(stderr, "invalid id, aborted\n");
				break;
			}
			strcpy(c.id, buf);
			execcmd(&c);
		case 's': // set filter
			printf("filter: ");
			if (fgets(buf, MAXLINE, stdin) == NULL)
				return;
			n = strlen(buf);
			if (buf[n - 1] != '\n') {
				fprintf(stderr, "filter too long\n");
				break;
			}
			buf[n - 1] = '\0';
			selectfiles(&c, buf);
			if (c.nselected <= 1) {
				fprintf(stderr, "no match\n");
				c.nselected = c.nvalid;
			} else {
				fprintf(stderr, "%d files selected\n", c.nselected - 1);
			}
			break;
		case 'z': // compres files
			setenv("COMPRESS_PASSWORD", secret, 1);
			c.files[0] = "/bin/compress"; // TODO: install it
			execcmd(&c);
			break;
		case 'q': // quit
			return;
		default:
			fprintf(stderr, "unknown command\n");
			break;
		}
	}
}

// NO INTENTIONAL BUGS AFTER THIS POINT
void sigchld(int signo)
{
	int status;
	pid_t pid = waitpid(-1, &status, WNOHANG);
	if (pid > 0) {
		printf("child %ld: ", (long)pid);
		if (WIFEXITED(status)) {
			printf("exited with status: %d\n", WEXITSTATUS(status));
		} else {
			printf("%s\n", strsignal(WTERMSIG(status)));
		}
	}
}

int main()
{
	int lstn;
	int enable;
	struct sockaddr_in lstn_addr;

	setvbuf(stdout, NULL, _IONBF, 0);
	setvbuf(stderr, NULL, _IONBF, 0);

	lstn = socket(AF_INET, SOCK_STREAM, 0);
	if (lstn < 0) {
		perror("socket");
		return 1;
	}
	enable = 1;
	if (setsockopt(lstn, SOL_SOCKET, SO_REUSEADDR, &enable, sizeof(enable)) < 0) {
		perror("setsockopt");
		return 1;
	}
	bzero(&lstn_addr, sizeof(lstn_addr));

	lstn_addr.sin_family = AF_INET;
	lstn_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	lstn_addr.sin_port = htons(PORT);

	if (bind(lstn, (struct sockaddr *)&lstn_addr, sizeof(lstn_addr)) < 0) {
		perror("bind");
		return 1;
	}

	if (listen(lstn, 10) < 0) {
		perror("listen");
		return 1;
	}
	printf("Listening on port %d\n", PORT);

	signal(SIGCHLD, sigchld);

	for (;;) {
		int con = accept(lstn, NULL, NULL);
		if (con < 0) {
			perror("accept");
			return 1;
		}

		switch (fork()) {
		case -1:
			perror("fork");
			return 1;
		case 0:
			printf("New connection, child %d\n", getpid());

			close(0);
			if (dup(con) < 0)
				goto error;
			close(1);
			if (dup(con) < 0)
				goto error;
			close(2);
			if (dup(con) < 0)
				goto error;
			child();
			exit(0);
			break;
		default:
			close(con);
			break;
		}
	}
	return 0;

error:
	fprintf(stderr, "dup() failed\n");
	exit(1);
}

void readfiles(struct cmd *c)
{
	DIR *dir;
	struct dirent *ent;
	char *ptr;
	size_t len, rem;

	dir = opendir(".");
	if (dir == NULL) {
		perror(".");
		exit(1);
	}

	ptr = filenames;
	rem = BUFSZ;
	c->nvalid = 1;
	while ( c->nvalid < MAXFILES && (ent = readdir(dir)) ) {
		len = strlen(ent->d_name) + 1;
		if (ent->d_name[0] == '.')
			continue;
		if (len > rem)
			break;
		c->files[c->nvalid] = ptr;
		strcpy(ptr, ent->d_name);
		ptr += len;
		rem -= len;
		c->nvalid++;
	}
	c->nselected = c->nvalid;
	closedir(dir);
}

void selectfiles(struct cmd *c, const char *filter)
{
	int cur;

	// the first entry actually contains the command name
	// and is always selected
	c->nselected = 1;
	for (cur = 1; cur < c->nvalid; cur++) {
		if (strstr(c->files[cur], filter)) {
			if (cur != c->nselected) {
				char *tmp = c->files[c->nselected];
				c->files[c->nselected] = c->files[cur];
				c->files[cur] = tmp;
			}
			c->nselected++;
		}
	}
}

void fillsecret(char *secret)
{
	FILE *fp;
	fp = fopen("../flag.txt", "r");
	if (fp == NULL) {
		perror("flag.txt");
		exit(1);
	}
	if (fgets(secret, MAXLINE, fp) == NULL) {
		fprintf(stderr, "unable to load the secret\n");
		exit(1);
	}
	fclose(fp);
}
